/*
 * dumpbw.c - Interpret a file as a grayscale image, whether it really
 *	is or not.  Useful for finding graphics within other files.
 *	This is not intended for copyright infringement.
 *
 * Copyright 2000, George Greer <greerga@fftactics.org>
 *
 * Licensed under the GNU General Public License (GPL)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *   or visit ftp://ftp.gnu.org/pub/gnu/COPYING via the Internet.
 */

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <gd.h>

#include "mmaplib.h"

#define FALSE	0
#define TRUE	(!FALSE)

static const char *program = __FILE__ ": Raw 16-bit color or 8-bit mono graphic translator.";
static const char *version = "v0.6 (Build: "__DATE__" "__TIME__")";
static const char *copyright = "Copyright 2000, George Greer <greerga@fftactics.org>";
static const char *license = "Licensed under GNU General Public License (GPL) v2 or later.";

int print_usage(const char *progname)
{
  printf("Usage: %s [-j offset] [-l number] [-c] [-o outputfile] inputfile\n%s --version\n%s --help\n", progname, progname, progname);
  return 1;
}

int print_version(const char *progname)
{
  printf("%s\n\t%s\n\t%s\n\n%s: version %s\n", program, copyright, license, progname, version);
  return 0;
}

struct option /* {
  const char *name;
  int has_arg;
  int *flag;
  int val;
} */ long_options[] = {
  { "help",     FALSE, NULL, 0 },
  { "version",  FALSE, NULL, 1 },
  { 0, 0, 0, 0 }
};

struct options_s {
  const char *outfile;
  long skipdata;
  unsigned int printlen;
  unsigned int blocksize;
  unsigned int color : 1;
  unsigned int xscale;
  unsigned int yscale;
};

union palette_u {
  struct {
    unsigned int r : 5;
    unsigned int g : 5;
    unsigned int b : 5;
    unsigned int t : 1;
  } bits;
  unsigned short int bytes;
};

/*
 * 16-bit color, 5551, Red-Green-Blue-Transparent
 */
struct palette_s {
  unsigned char r;
  unsigned char g;
  unsigned char b;
  unsigned char t : 1;
};

void make_color(FILE *output, gdImagePtr pic, const unsigned char *data, size_t length, struct options_s *options);
void make_bw(FILE *output, gdImagePtr pic, const unsigned char *data, size_t length, struct options_s *options);
void dumpbw(const unsigned char *data, size_t length, struct options_s *options);

int main(int argc, char *argv[])
{
  struct options_s options;
  int ch, zoom;

  memset(&options, 0, sizeof(options));
  options.blocksize = 128;
  options.xscale = 1;
  options.yscale = 1;

  while ((ch = getopt_long(argc, argv, ":cj:l:o:z:", long_options, NULL)) >= 0) {
    switch (ch) {
    case 0:
      return print_usage(argv[0]);
    case 1:
      return print_version(argv[0]);
    case 'c':
      options.color = TRUE;
      break;
    case 'l':
      options.blocksize = atoi(optarg);
      break;
    case 'j':
      if (sscanf(optarg, "%ld:%d", &options.skipdata, &options.printlen) != 2)
        if (sscanf(optarg, "%ld-%d", &options.skipdata, &options.printlen) == 2)
          options.printlen -= options.skipdata;
      break;
    case 'o':
      options.outfile = optarg;
      break;
    case 'z':
      sscanf(optarg, "%d:%d:%d", &zoom, &options.xscale, &options.yscale);
      options.xscale *= zoom;
      options.yscale *= zoom;
      error("zoom not yet supported");
      break;
    case ':':
      fprintf(stderr, "Option -%c requires an extra argument.\n", optopt);
      return 1;
    case '?':
      fprintf(stderr, "Unknown option -%c.\n", optopt);
      return 1;
    default:
      fputs("Unknown error condition.\n", stderr);
      return 1;
    }
  }

  if (argc == optind)
    fputs("Program requires a filename to translate.\nTry '--help' for more information.", stderr);
  else if (options.blocksize == 0)
    fputs("invalid block length 0", stderr);
  else if (options.xscale + options.yscale == 0)
    fputs("invalid zoom level 0", stderr);
  else {
    void *handle;
    if (!(handle = mapopen(argv[optind])))
      return 1;
    dumpbw(mapdata(handle), maplength(handle), &options);
    mapclose(handle);
    return 0;
  }
  return 1;
}

void dumpbw(const unsigned char *data, size_t length, struct options_s *options)
{
  const unsigned char *start;
  const char *filename;
  size_t span;
  FILE *output;
  gdImagePtr pic;
  int ysize;

  start = options->skipdata >= length ? data : data + options->skipdata;

  if (options->printlen && start + options->printlen <= data + length)
    span = options->printlen;
  else
    span = length - (start - data);

  filename = options->outfile ? options->outfile : "dumpbw-output.gif";
  if (!(output = fopen(filename, "wb")))
    error(filename);

  if (options->color)
    ysize = ((span / 2) / options->blocksize) + 0.5;
  else
    ysize = (span / options->blocksize) + 0.5;

  pic = gdImageCreate(options->blocksize, ysize);

  if (options->color)
    make_color(output, pic, start, span, options);
  else
    make_bw(output, pic, start, span, options);

  gdImageGif(pic, output);
  gdImageDestroy(pic);
  fclose(output);
}

void make_color(FILE *output, gdImagePtr pic, const unsigned char *data, size_t length, struct options_s *options)
{
  int x1, y1, itr, clr;
  union palette_u pixel;
  struct palette_s dot;

  for (y1 = itr = 0; itr < length; y1++)
    for (x1 = 0; x1 < options->blocksize && itr < length; x1++, itr += 2) {
      memcpy(&pixel.bytes, data + itr, sizeof(short));

      dot.r = pixel.bits.r * 255 / 0x1F;
      dot.g = pixel.bits.g * 255 / 0x1F;
      dot.b = pixel.bits.b * 255 / 0x1F;

      if ((clr = gdImageColorExact(pic, dot.r, dot.g, dot.b)) < 0)
        if ((clr = gdImageColorAllocate(pic, dot.r, dot.g, dot.b)) < 0)
          if ((clr = gdImageColorClosest(pic, dot.r, dot.g, dot.b)) < 0)
            fprintf(stderr, "%d: cannot get color %d/%d/%d", itr, dot.r, dot.g, dot.b);

      gdImageFilledRectangle(pic, x1, y1, x1 + 1, y1, clr);
    }
}

void make_bw(FILE *output, gdImagePtr pic, const unsigned char *data, size_t length, struct options_s *options)
{
  int x1, y1, itr;

  for (itr = 0; itr < 256; itr++)
    if (gdImageColorAllocate(pic, itr, itr, itr) != itr)
      error("GD palette mismatch");

  for (y1 = itr = 0; itr < length; y1++)
    for (x1 = 0; x1 < options->blocksize && itr < length; x1++, itr++)
      gdImageFilledRectangle(pic, x1, y1, x1 + 2, y1 + 1, data[itr]);
}
